package lcsb.mapviewer.converter.model.celldesigner;

import lcsb.mapviewer.commands.CreateHierarchyCommand;
import lcsb.mapviewer.converter.ComplexZipConverter;
import lcsb.mapviewer.converter.ComplexZipConverterParams;
import lcsb.mapviewer.converter.ProjectFactory;
import lcsb.mapviewer.converter.zip.ZipEntryFile;
import lcsb.mapviewer.converter.zip.ZipEntryFileFactory;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.ProjectComparator;
import lcsb.mapviewer.model.map.compartment.PathwayCompartment;
import lcsb.mapviewer.model.map.layout.graphics.Glyph;
import lcsb.mapviewer.model.map.layout.graphics.Layer;
import lcsb.mapviewer.model.map.layout.graphics.LayerImage;
import lcsb.mapviewer.model.map.layout.graphics.LayerText;
import lcsb.mapviewer.model.map.model.ElementSubmodelConnection;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.map.model.ModelSubmodelConnection;
import lcsb.mapviewer.model.map.model.SubmodelType;
import lcsb.mapviewer.model.map.species.Protein;
import lcsb.mapviewer.model.overlay.DataOverlay;
import lcsb.mapviewer.model.overlay.DataOverlayEntry;
import lcsb.mapviewer.model.overlay.GenericDataOverlayEntry;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotEquals;

public class ProjectExportTest extends CellDesignerTestFunctions {

  private final ProjectComparator projectComparator = new ProjectComparator();

  private final ZipEntryFileFactory zefFactory = new ZipEntryFileFactory();

  private ComplexZipConverter parser;
  private ProjectFactory projectFactory;

  @Before
  public void setUp() throws Exception {
    parser = new ComplexZipConverter(CellDesignerXmlParser.class);
    projectFactory = new ProjectFactory(parser);
  }

  @After
  public void tearDown() throws Exception {
  }

  @Test
  public void testBasic() throws Exception {
    Project project = createProject();
    Model model = createEmptyModel();
    project.addModel(model);

    testSerializationOverZip(project);
  }

  @Test
  public void testWithManyModels() throws Exception {
    Project project = createProject();
    Model model = createEmptyModel();
    project.addModel(model);

    Model model2 = createEmptyModel();
    model2.setWidth(260);
    model2.addElement(createProtein());
    project.addModel(model2);
    model.addSubmodelConnection(new ModelSubmodelConnection(model2, SubmodelType.UNKNOWN, model2.getName()));

    testSerializationOverZip(project);
  }

  @Test
  public void testLinkToSubmodel() throws Exception {
    Project project = createProject();
    Model model = createEmptyModel();
    model.setWidth(260);
    Protein protein = createProtein();
    model.addElement(protein);
    project.addModel(model);

    Model model2 = createEmptyModel();
    project.addModel(model2);
    model.addSubmodelConnection(new ModelSubmodelConnection(model2, SubmodelType.UNKNOWN, model2.getName()));
    protein.setSubmodel(new ElementSubmodelConnection(model2, SubmodelType.DOWNSTREAM_TARGETS, model2.getName()));

    testSerializationOverZip(project);
  }

  @Test
  public void testImmediateLink() throws Exception {
    Project project = createProject();
    Model model = createEmptyModel();
    model.setWidth(260);
    Protein protein = createProtein();
    protein.setImmediateLink("https://uni.lu/");
    model.addElement(protein);
    project.addModel(model);

    testSerializationOverZip(project);
  }

  @Test
  public void testGlyph() throws Exception {
    Glyph glyph = createGlyph();
    Project project = createProject();
    Model model = createEmptyModel();
    model.setWidth(260);
    Protein protein = createProtein();
    protein.setGlyph(glyph);
    model.addElement(protein);
    project.addModel(model);
    project.addGlyph(glyph);

    testSerializationOverZip(project);
  }

  @Test
  public void testDataOverlay() throws Exception {
    Project project = createProject();
    Model model = createEmptyModel();
    model.setWidth(260);
    Protein protein = createProtein();
    model.addElement(protein);
    project.addModel(model);

    DataOverlay overlay = createDataOverlay();
    overlay.setPublic(true);
    overlay.addEntry(createDataOverlayEntry());
    overlay.addEntry(createDataOverlayEntry());

    project.addDataOverlay(overlay);

    testSerializationOverZip(project);
  }

  @Test
  public void testPrivateDataOverlay() throws Exception {
    Project project = createProject();
    Model model = createEmptyModel();
    model.setWidth(260);
    Protein protein = createProtein();
    model.addElement(protein);
    project.addModel(model);

    DataOverlay overlay = createDataOverlay();
    overlay.setPublic(false);
    DataOverlayEntry entry = new GenericDataOverlayEntry();
    entry.setName(protein.getName());
    entry.setValue(faker.number().randomDouble(2, -1, 1));

    project.addDataOverlay(overlay);

    testSerializationOverZip(project, false);
  }

  private void testSerializationOverZip(final Project project) throws Exception {
    testSerializationOverZip(project, true);
  }

  private void testSerializationOverZip(final Project project, final boolean shouldBeTheSame) throws Exception {
    File tempFile = File.createTempFile("CD-", ".zip");
    try (FileOutputStream outputStream = new FileOutputStream(tempFile)) {
      byte[] data = projectFactory.project2zip(project);
      outputStream.write(data);
    }
    final ComplexZipConverterParams complexParams = createDefaultParams(tempFile);
    Project project2 = projectFactory.create(complexParams);

    for (ModelData model : project.getModels()) {
      clearModelFromZ(model);
    }
    for (ModelData model : project2.getModels()) {
      new CreateHierarchyCommand(model.getModel(), 10, 1000).execute();
      clearModelFromZ(model);
    }

    if (shouldBeTheSame) {
      assertEquals(0, projectComparator.compare(project, project2));
    } else {
      assertNotEquals(0, projectComparator.compare(project, project2));
    }
  }

  private ComplexZipConverterParams createDefaultParams(final File tempFile) throws IOException {
    final ComplexZipConverterParams complexParams;
    complexParams = new ComplexZipConverterParams().zipFile(new ZipFile(tempFile));
    complexParams.visualizationDir(faker.name().username());

    ZipFile zipFile = new ZipFile(tempFile.getAbsolutePath());
    Enumeration<? extends ZipEntry> entries = zipFile.entries();
    while (entries.hasMoreElements()) {
      ZipEntry entry = entries.nextElement();
      ZipEntryFile e = zefFactory.createZipEntryFile(entry, zipFile);
      complexParams.entry(e);
    }
    zipFile.close();
    return complexParams;
  }

  @Test
  public void testLayerWithText() throws Exception {
    Project project = createProject();
    Model model = createEmptyModel();
    model.setWidth(745);
    model.setHeight(346);
    Layer layer = createLayer();
    LayerText layerText = createText();
    layer.addLayerText(layerText);
    model.addLayer(layer);

    project.addModel(model);

    testSerializationOverZip(project);
  }

  @Test
  public void testLayerWithImage() throws Exception {
    Project project = createProject();
    Model model = createEmptyModel();
    model.setWidth(745);
    model.setHeight(346);
    Layer layer = createLayer();
    LayerImage image = createImage();
    layer.addLayerImage(image);
    model.addLayer(layer);

    project.addModel(model);
    project.addGlyph(image.getGlyph());

    testSerializationOverZip(project);
  }

  @Test
  public void testWithPathway() throws Exception {
    Project project = createProject();
    Model model = createEmptyModel();
    model.setWidth(126);
    model.setHeight(148);
    PathwayCompartment pathway = createPathway();
    model.addElement(pathway);

    project.addModel(model);

    testSerializationOverZip(project);
  }

}
