package lcsb.mapviewer.converter;

import lcsb.mapviewer.common.geometry.ColorParser;
import lcsb.mapviewer.converter.zip.ZipEntryFileFactory;
import lcsb.mapviewer.model.map.BioEntity;
import lcsb.mapviewer.model.map.species.AntisenseRna;
import lcsb.mapviewer.model.map.species.Complex;
import lcsb.mapviewer.model.map.species.Degraded;
import lcsb.mapviewer.model.map.species.Drug;
import lcsb.mapviewer.model.map.species.Gene;
import lcsb.mapviewer.model.map.species.Ion;
import lcsb.mapviewer.model.map.species.Phenotype;
import lcsb.mapviewer.model.map.species.Protein;
import lcsb.mapviewer.model.map.species.Rna;
import lcsb.mapviewer.model.map.species.SimpleMolecule;
import lcsb.mapviewer.model.map.species.Unknown;
import lcsb.mapviewer.model.overlay.DataOverlay;
import lcsb.mapviewer.model.overlay.DataOverlayEntry;
import lcsb.mapviewer.model.overlay.DataOverlayType;
import org.apache.commons.lang3.StringUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class ColorSchemaWriter {

  private static final Map<String, Class<? extends BioEntity>> speciesMapping;

  static {
    speciesMapping = new HashMap<>();
    speciesMapping.put("protein", Protein.class);
    speciesMapping.put("gene", Gene.class);
    speciesMapping.put("complex", Complex.class);
    speciesMapping.put("simple_molecule", SimpleMolecule.class);
    speciesMapping.put("ion", Ion.class);
    speciesMapping.put("phenotype", Phenotype.class);
    speciesMapping.put("drug", Drug.class);
    speciesMapping.put("rna", Rna.class);
    speciesMapping.put("antisense_rna", AntisenseRna.class);
    speciesMapping.put("unknown", Unknown.class);
    speciesMapping.put("degraded", Degraded.class);
  }

  /**
   * Object that parses colors from string.
   */
  private final ColorParser colorParser = new ColorParser();

  /**
   * Default class logger.
   */
  private final Logger logger = LogManager.getLogger();

  public String entriesToStringTable(final Set<DataOverlayEntry> entries) {
    List<ColorSchemaColumn> columns = new ArrayList<>();
    for (ColorSchemaColumn column : ColorSchemaColumn.values()) {
      if (column.getTypes().contains(DataOverlayType.GENERIC)) {
        columns.add(column);
      }
    }

    StringBuilder builder = new StringBuilder();
    builder.append(getHeader(columns));
    for (DataOverlayEntry entry : entries) {
      builder.append("\n").append(getRow(entry, columns));
    }

    return builder.toString();
  }

  private String getRow(final DataOverlayEntry entry, final List<ColorSchemaColumn> columns) {
    StringBuilder row = new StringBuilder();
    int counter = 0;
    for (ColorSchemaColumn column : columns) {
      if (counter > 0) {
        row.append("\t");
      }
      counter++;
      String data = "";
      switch (column) {
        case COLOR:
          if (entry.getColor() != null) {
            data = new ColorParser().colorToHtml(entry.getColor());
          }
          break;
        case NAME:
          if (entry.getName() != null) {
            data = entry.getName();
          }
          break;
        case ELEMENT_IDENTIFIER:
          if (entry.getElementId() != null) {
            data = entry.getElementId();
          }
          break;
        case MAP_NAME:
          if (entry.getModelName() != null) {
            data = entry.getModelName();
          }
          break;
        case LINE_WIDTH:
          if (entry.getLineWidth() != null) {
            data = entry.getLineWidth().toString();
          }
          break;
        case VALUE:
          if (entry.getValue() != null) {
            data = entry.getValue().toString();
          }
          break;
        case REVERSE_REACTION:
          if (entry.getReverseReaction() != null) {
            data = entry.getReverseReaction().toString();
          }
          break;
        case COMPARTMENT:
          data = StringUtils.join(entry.getCompartments(), ",");
          break;
        case DESCRIPTION:
          if (entry.getDescription() != null) {
            data = entry.getDescription().replace("\t", " ").replace("\n", " ");
          }
          break;
        default:
          data = "";
          break;
      }
      row.append(data);
    }
    return row.toString();
  }

  private String getHeader(final List<ColorSchemaColumn> columns) {
    StringBuilder header = new StringBuilder();
    for (ColorSchemaColumn column : columns) {
      if (header.length() > 0) {
        header.append("\t");
      }
      header.append(column.toString());
    }
    return header.toString();
  }

  public String overlayToString(final DataOverlay overlay) {
    String metaData = overlayMetaData(overlay);

    String content = metaData + entriesToStringTable(overlay.getEntries());
    return content;
  }

  private String overlayMetaData(final DataOverlay overlay) {
    StringBuilder metaData = new StringBuilder();
    if (overlay.getName() != null) {
      metaData.append("#" + ZipEntryFileFactory.LAYOUT_HEADER_PARAM_NAME + "=" + overlay.getName() + "\n");
    }
    if (overlay.getDescription() != null) {
      metaData.append("#" + ZipEntryFileFactory.LAYOUT_HEADER_PARAM_DESCRIPTION + "=" + overlay.getDescription().replace("\n", " ") + "\n");
    }

    return metaData.toString();
  }
}
