package lcsb.mapviewer.converter;

import lcsb.mapviewer.commands.layout.ApplySimpleLayoutModelCommand;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.converter.zip.GlyphZipEntryFile;
import lcsb.mapviewer.converter.zip.ImageZipEntryFile;
import lcsb.mapviewer.converter.zip.LayoutZipEntryFile;
import lcsb.mapviewer.converter.zip.ZipEntryFile;
import lcsb.mapviewer.converter.zip.ZipEntryFileFactory;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.cache.UploadedFileEntry;
import lcsb.mapviewer.model.map.InconsistentModelException;
import lcsb.mapviewer.model.map.compartment.Compartment;
import lcsb.mapviewer.model.map.compartment.SquareCompartment;
import lcsb.mapviewer.model.map.layout.graphics.Glyph;
import lcsb.mapviewer.model.map.layout.graphics.Layer;
import lcsb.mapviewer.model.map.layout.graphics.LayerImage;
import lcsb.mapviewer.model.map.layout.graphics.LayerText;
import lcsb.mapviewer.model.map.model.ElementSubmodelConnection;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.map.model.ModelFullIndexed;
import lcsb.mapviewer.model.map.model.ModelSubmodelConnection;
import lcsb.mapviewer.model.map.reaction.Product;
import lcsb.mapviewer.model.map.reaction.Reactant;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.reaction.type.TransportReaction;
import lcsb.mapviewer.model.map.species.Complex;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.GenericProtein;
import lcsb.mapviewer.model.map.species.Species;
import lcsb.mapviewer.model.overlay.DataOverlay;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;

public class ProjectFactory {

  /**
   * Default class logger.
   */
  @SuppressWarnings("unused")
  private final Logger logger = LogManager.getLogger();

  private final ComplexZipConverter converter;

  private final GlyphParser glyphParser = new GlyphParser();

  public ProjectFactory(final ComplexZipConverter converter) {
    this.converter = converter;
  }

  private static void addModelFileToZip(final Model model, final String filename, final Converter converter, final ZipOutputStream zos)
      throws IOException, InconsistentModelException, ConverterException {
    ZipEntry entry = new ZipEntry(filename);
    zos.putNextEntry(entry);
    String topModelContent = converter.model2String(model);
    zos.write(topModelContent.getBytes());
    zos.closeEntry();
  }

  public Project create(final ComplexZipConverterParams params) throws InvalidInputDataExecption, ConverterException {
    return create(params, new Project());
  }

  public Project create(final ComplexZipConverterParams params, final Project project)
      throws InvalidInputDataExecption, ConverterException {
    try {
      ZipFile zipFile = params.getZipFile();
      Enumeration<? extends ZipEntry> entries;

      entries = zipFile.entries();
      List<ImageZipEntryFile> imageEntries = new ArrayList<>();
      int overlayOrder = 1;
      while (entries.hasMoreElements()) {
        ZipEntry entry = entries.nextElement();
        if (!entry.isDirectory()) {
          ZipEntryFile zef = params.getEntry(entry.getName());
          if (zef instanceof ImageZipEntryFile) {
            imageEntries.add((ImageZipEntryFile) zef);
          } else if (zef instanceof GlyphZipEntryFile) {
            project.addGlyph(glyphParser.parseGlyph((GlyphZipEntryFile) zef, zipFile));
          } else if (zef instanceof LayoutZipEntryFile) {
            project.addDataOverlay(
                converter.layoutZipEntryFileToLayout(params, zipFile, entry, (LayoutZipEntryFile) zef, overlayOrder++));
          }
        }
      }
      converter.createModel(params, project);


      if (!imageEntries.isEmpty()) {
        OverviewParser parser = new OverviewParser();
        project.addOverviewImages(parser.parseOverviewLinks(project.getModels(), imageEntries, params.getVisualizationDir(), zipFile));
      }
      if (!project.getGlyphs().isEmpty()) {
        assignGlyphsToElements(project);
      }
      return project;
    } catch (final IOException e) {
      throw new InvalidArgumentException(e);
    } catch (final InvalidCoordinatesFile e) {
      throw new InvalidInputDataExecption("Invalid coordinates file. " + e.getMessage(), e);
    }
  }

  private void assignGlyphsToElements(final Project project) throws InvalidGlyphFile {
    Set<ModelData> models = new HashSet<>(project.getModels());

    for (final ModelData model : project.getModels()) {
      for (final ModelSubmodelConnection connection : model.getSubmodels()) {
        models.add(connection.getSubmodel());
      }
    }
    for (final ModelData model : models) {
      for (final Element element : model.getElements()) {
        Glyph glyph = extractGlyph(project, element.getNotes());
        if (glyph != null) {
          element.setNotes(removeGlyph(element.getNotes()));
          element.setGlyph(glyph);
        }
      }
      for (final Layer layer : model.getLayers()) {
        for (final LayerText text : layer.getTexts()) {
          Glyph glyph = extractGlyph(project, text.getNotes());
          if (glyph != null) {
            text.setNotes(removeGlyph(text.getNotes()));
            text.setGlyph(glyph);
          }
        }
      }
    }
  }

  String removeGlyph(final String notes) {
    String[] lines = notes.split("[\n\r]+");
    StringBuilder result = new StringBuilder();
    for (final String line : lines) {
      if (!line.startsWith(ComplexZipConverter.GLYPH_PREFIX)) {
        result.append(line).append("\n");
      }
    }
    return result.toString();
  }

  Glyph extractGlyph(final Project project, final String notes) throws InvalidGlyphFile {
    String[] lines = notes.split("[\n\r]+");
    for (final String line : lines) {
      if (line.startsWith(ComplexZipConverter.GLYPH_PREFIX)) {
        String glyphString = line.replace(ComplexZipConverter.GLYPH_PREFIX, "").trim().toLowerCase();
        for (final Glyph glyph : project.getGlyphs()) {
          if (glyph.getFile().getOriginalFileName().toLowerCase().equalsIgnoreCase(glyphString)) {
            return glyph;
          }
        }
        throw new InvalidGlyphFile("Glyph file doesn't exist: " + glyphString);
      }
    }
    return null;
  }

  public byte[] project2zip(final Project project) throws ConverterException, InconsistentModelException {
    Converter converter = this.converter.createConverterInstance();

    ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
    try (ZipOutputStream zos = new ZipOutputStream(byteArrayOutputStream)) {
      addModelFileToZip(project.getTopModel(), project.getTopModel().getName() + "." + converter.getFileExtensions().get(0), converter, zos);
      for (ModelData model : project.getModels()) {
        if (model != project.getTopModelData()) {
          addModelFileToZip(model.getModel(),
              ZipEntryFileFactory.SUBMODEL_DIRECTORY + model.getName() + "." + converter.getFileExtensions().get(0), converter,
              zos);
        }
      }
      Model mapping = createMappingModel(project.getModels());

      addModelFileToZip(mapping, ZipEntryFileFactory.SUBMODEL_DIRECTORY + "mapping." + converter.getFileExtensions().get(0), converter, zos);

      for (ModelData model : project.getModels()) {
        addGlyphsToZip(model.getModel(), zos);
      }

      for (DataOverlay overlay : project.getDataOverlays()) {
        if (overlay.isPublic()) {
          addDataOverlayToZip(overlay, zos);
        }
      }

    } catch (IOException ioe) {
      throw new ConverterException(ioe);
    }
    return byteArrayOutputStream.toByteArray();
  }

  private void addDataOverlayToZip(final DataOverlay overlay, final ZipOutputStream zos) throws IOException {
    ColorSchemaWriter writer = new ColorSchemaWriter();
    ZipEntry entry = new ZipEntry(ZipEntryFileFactory.LAYOUT_DIRECTORY + overlay.getId() + "_" + overlay.getName());
    zos.putNextEntry(entry);
    zos.write(writer.overlayToString(overlay).getBytes());
    zos.closeEntry();
  }

  private void addGlyphsToZip(final Model model, final ZipOutputStream zos) throws IOException {
    for (Element element : model.getElements()) {
      if (element.getGlyph() != null) {
        UploadedFileEntry uploadedFileEntry = element.getGlyph().getFile();
        ZipEntry entry = new ZipEntry(ZipEntryFileFactory.GLYPHS_DIRECTORY + new File(uploadedFileEntry.getOriginalFileName()).getName());
        zos.putNextEntry(entry);
        zos.write(uploadedFileEntry.getFileContent());
        zos.closeEntry();
      }
    }
    for (Layer layer : model.getLayers()) {
      for (LayerImage image : layer.getImages()) {
        if (image.getGlyph() != null) {
          UploadedFileEntry uploadedFileEntry = image.getGlyph().getFile();
          ZipEntry entry = new ZipEntry(ZipEntryFileFactory.GLYPHS_DIRECTORY + new File(uploadedFileEntry.getOriginalFileName()).getName());
          zos.putNextEntry(entry);
          zos.write(uploadedFileEntry.getFileContent());
          zos.closeEntry();
        }
      }
    }
  }

  private int idCounter = 0;

  private Model createMappingModel(final Set<ModelData> models) {
    Model mapping = new ModelFullIndexed(null);
    for (ModelData parentModel : models) {
      for (Element parentElement : parentModel.getElements()) {
        ElementSubmodelConnection connection = parentElement.getSubmodel();
        if (connection != null) {
          ModelData childModel = connection.getSubmodel();

          Complex mappingParentElement = getComplexForConnection(parentModel.getName(), mapping);

          Species sourceElement = getComplexChildForConnection(parentElement.getElementId(), mappingParentElement);

          Complex mappingChildElement = getComplexForConnection(childModel.getName(), mapping);

          Species targetElement;
          if (connection.getToElement() != null) {
            targetElement = getComplexChildForConnection(connection.getToElement().getElementId(), mappingChildElement);
          } else {
            targetElement = mappingChildElement;
          }

          Reaction reaction = new TransportReaction("re" + idCounter++);
          reaction.addReactant(new Reactant(sourceElement));
          reaction.addProduct(new Product(targetElement));
          mapping.addReaction(reaction);
        }
        if (parentElement.getImmediateLink() != null) {
          Compartment mappingParentCompartment = getCompartmentForConnection(parentModel.getName(), mapping);

          Species sourceElement = getCompartmentChildForConnection(parentElement.getElementId(), mappingParentCompartment);
          String notes = sourceElement.getNotes();
          notes += "\n" + ComplexZipConverter.IMMEDIATE_LINK_PREFIX + parentElement.getImmediateLink();
          sourceElement.setNotes(notes);
        }
        if (parentElement.getGlyph() != null) {
          UploadedFileEntry uploadedFileEntry = parentElement.getGlyph().getFile();
          Compartment mappingParentCompartment = getCompartmentForConnection(parentModel.getName(), mapping);

          Species sourceElement = getCompartmentChildForConnection(parentElement.getElementId(), mappingParentCompartment);
          String notes = sourceElement.getNotes();
          notes += "\n" + ComplexZipConverter.GLYPH_PREFIX + "glyphs/" + new File(uploadedFileEntry.getOriginalFileName()).getName();
          sourceElement.setNotes(notes);
        }
      }
    }
    try {
      new ApplySimpleLayoutModelCommand(mapping).execute();
    } catch (Exception e) {
      logger.error("Problem with preparing mapping file", e);
    }
    return mapping;
  }

  private Species getComplexChildForConnection(final String elementId, final Complex mappingParentElement) {
    Species sourceElement = null;
    for (Species child : mappingParentElement.getAllChildren()) {
      if (child.getName().equals(elementId)) {
        sourceElement = child;
      }
    }

    if (sourceElement == null) {
      sourceElement = new GenericProtein("sa" + idCounter++);
      sourceElement.setName(elementId);
      mappingParentElement.addSpecies(sourceElement);
      mappingParentElement.getModel().addElement(sourceElement);
    }
    return sourceElement;
  }

  private Species getCompartmentChildForConnection(final String elementId, final Compartment mappingParentElement) {
    Species sourceElement = null;
    for (Element child : mappingParentElement.getElements()) {
      if (child instanceof Species && child.getName().equals(elementId)) {
        sourceElement = (Species) child;
      }
    }

    if (sourceElement == null) {
      sourceElement = new GenericProtein("sa" + idCounter++);
      sourceElement.setName(elementId);
      mappingParentElement.addElement(sourceElement);
      mappingParentElement.getModel().addElement(sourceElement);
    }
    return sourceElement;
  }

  private Complex getComplexForConnection(final String modelName, final Model mapping) {
    List<Element> mappingParentElements = mapping.getElementsByName(modelName);

    Complex mappingParentElement = null;
    for (Element mappingElement : mappingParentElements) {
      if (mappingElement instanceof Complex) {
        mappingParentElement = (Complex) mappingElement;
      }
    }
    if (mappingParentElement == null) {
      mappingParentElement = new Complex("ca" + idCounter++);
      mappingParentElement.setName(modelName);
      mapping.addElement(mappingParentElement);
    }
    return mappingParentElement;
  }

  private Compartment getCompartmentForConnection(final String modelName, final Model mapping) {
    List<Element> mappingParentElements = mapping.getElementsByName(modelName);

    Compartment mappingParentElement = null;
    for (Element mappingElement : mappingParentElements) {
      if (mappingElement instanceof Compartment) {
        mappingParentElement = (Compartment) mappingElement;
      }
    }
    if (mappingParentElement == null) {
      mappingParentElement = new SquareCompartment("ca" + idCounter++);
      mappingParentElement.setName(modelName);
      mapping.addElement(mappingParentElement);
    }
    return mappingParentElement;
  }
}
