package lcsb.mapviewer.commands;

import lcsb.mapviewer.model.map.BioEntity;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.model.ModelComparator;
import lcsb.mapviewer.model.map.model.ModelFullIndexed;
import lcsb.mapviewer.model.map.model.ModelSubmodelConnection;
import lcsb.mapviewer.model.map.model.SubmodelType;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.Gene;
import lcsb.mapviewer.model.map.species.GenericProtein;
import lcsb.mapviewer.model.map.species.Protein;
import lcsb.mapviewer.model.map.species.field.TranscriptionSite;
import lcsb.mapviewer.model.overlay.DataOverlayEntry;
import lcsb.mapviewer.model.overlay.GenericDataOverlayEntry;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

public class ColorModelCommandTest extends CommandTestFunctions {

  private final ColorExtractor colorExtractor = new ColorExtractor(Color.RED, Color.GREEN, Color.BLUE, Color.WHITE);

  @Before
  public void setUp() throws Exception {
  }

  @After
  public void tearDown() throws Exception {
  }

  @Test
  public void testAliasMatchWithInvalidElementId() {
    GenericDataOverlayEntry colorSchema = new GenericDataOverlayEntry();
    colorSchema.setName(null);
    colorSchema.setElementId("1");

    BioEntity protein = new GenericProtein("id");
    protein.setName("test");

    List<DataOverlayEntry> schemas = new ArrayList<>();
    schemas.add(colorSchema);

    ColorModelCommand factory = new ColorModelCommand(new ModelFullIndexed(null), schemas, colorExtractor);

    assertFalse(factory.match(protein, colorSchema));

    colorSchema.setElementId(null);
    assertTrue(factory.match(protein, colorSchema));
  }

  @Test
  public void testSpeciesMatchWithProteinType() {
    GenericDataOverlayEntry colorSchema = new GenericDataOverlayEntry();
    colorSchema.setName("s1");
    colorSchema.addType(Protein.class);

    GenericProtein species = new GenericProtein("id");
    species.setName("s1");

    List<DataOverlayEntry> schemas = new ArrayList<>();
    schemas.add(colorSchema);

    ColorModelCommand factory = new ColorModelCommand(new ModelFullIndexed(null), schemas, colorExtractor);

    assertTrue(factory.match(species, colorSchema));
  }

  @Test
  public void testSpeciesMatchWithMiriamData() {
    GenericDataOverlayEntry colorSchema = new GenericDataOverlayEntry();
    colorSchema.setName("s1");
    colorSchema.addMiriamData(new MiriamData(MiriamType.HGNC_SYMBOL, "SNCA"));

    GenericDataOverlayEntry colorSchema2 = new GenericDataOverlayEntry();
    colorSchema2.setName("s1");
    colorSchema2.addMiriamData(new MiriamData(MiriamType.HGNC_SYMBOL, "PARK7"));

    GenericProtein species = new GenericProtein("id");
    species.setName("s1");
    species.addMiriamData(new MiriamData(MiriamType.HGNC_SYMBOL, "SNCA"));

    List<DataOverlayEntry> schemas = new ArrayList<>();
    schemas.add(colorSchema);

    ColorModelCommand factory = new ColorModelCommand(new ModelFullIndexed(null), schemas, colorExtractor);

    assertTrue(factory.match(species, colorSchema));
    assertFalse(factory.match(species, colorSchema2));
  }

  @Test
  public void testSpeciesMatchWithMiriamDataDifferentAnnotator() {
    GenericDataOverlayEntry colorSchema = new GenericDataOverlayEntry();
    colorSchema.setName("s1");
    colorSchema.addMiriamData(new MiriamData(MiriamType.HGNC_SYMBOL, "SNCA"));

    GenericProtein species = new GenericProtein("id");
    species.setName("s1");
    species.addMiriamData(new MiriamData(MiriamType.HGNC_SYMBOL, "SNCA", Object.class));

    List<DataOverlayEntry> schemas = new ArrayList<>();
    schemas.add(colorSchema);

    ColorModelCommand factory = new ColorModelCommand(new ModelFullIndexed(null), schemas, colorExtractor);

    assertTrue(factory.match(species, colorSchema));
  }

  @Test
  public void testReactionMatchWithProteinMiriamData() {
    GenericDataOverlayEntry colorSchema = new GenericDataOverlayEntry();
    colorSchema.addMiriamData(new MiriamData(MiriamType.HGNC_SYMBOL, "SNCA"));

    Reaction reaction = new Reaction("re");

    List<DataOverlayEntry> schemas = new ArrayList<>();
    schemas.add(colorSchema);

    ColorModelCommand factory = new ColorModelCommand(new ModelFullIndexed(null), schemas, colorExtractor);

    assertFalse(factory.match(reaction, colorSchema));
  }

  @Test
  public void testReactionMatchWithMiriamData() {
    GenericDataOverlayEntry colorSchema = new GenericDataOverlayEntry();
    colorSchema.addMiriamData(new MiriamData(MiriamType.PUBMED, "1234"));

    Reaction reaction = new Reaction("re");
    reaction.addMiriamData(new MiriamData(MiriamType.PUBMED, "1234"));

    List<DataOverlayEntry> schemas = new ArrayList<>();
    schemas.add(colorSchema);

    ColorModelCommand factory = new ColorModelCommand(new ModelFullIndexed(null), schemas, colorExtractor);

    assertTrue(factory.match(reaction, colorSchema));
  }


  @Test
  public void testGetModifiedElements() {
    Reaction reaction = new Reaction("re");
    reaction.addMiriamData(new MiriamData(MiriamType.PUBMED, "1234"));

    Model model = new ModelFullIndexed(null);
    model.addReaction(reaction);

    GenericDataOverlayEntry colorSchema = new GenericDataOverlayEntry();
    colorSchema.addMiriamData(new MiriamData(MiriamType.PUBMED, "1234"));

    List<DataOverlayEntry> schemas = new ArrayList<>();
    schemas.add(colorSchema);

    ColorModelCommand factory = new ColorModelCommand(model, schemas, colorExtractor);

    assertNotNull(factory.getModifiedElements().get(reaction));
  }

  @Test
  public void testApplyColorToReaction() {
    Model model = createSimpleModel();
    Protein protein1 = createProteinWithLayout();
    Protein protein2 = createProteinWithLayout();
    model.addElement(protein1);
    model.addElement(protein2);
    Reaction reaction = createReactionWithLayout(protein1, protein1);
    reaction.setIdReaction("re1");
    model.addReaction(reaction);

    DataOverlayEntry schema = new GenericDataOverlayEntry();
    schema.setElementId("re1");
    schema.setColor(Color.YELLOW);
    schema.setName(null);

    ColorModelCommand factory = new ColorModelCommand(model, new ArrayList<>(), colorExtractor);

    factory.applyColor(reaction, schema);

    assertEquals(Color.YELLOW, reaction.getLine().getColor());
  }

  @Test
  public void testColorTranscriptionFactor() {
    Model model = new ModelFullIndexed(null);

    Gene gene = createGene();
    TranscriptionSite site = new TranscriptionSite();
    site.setBorderColor(Color.YELLOW);
    gene.addTranscriptionSite(site);
    model.addElement(gene);

    ColorModelCommand factory = new ColorModelCommand(model, new ArrayList<>(), colorExtractor);
    factory.execute();

    assertEquals(Color.BLACK, site.getBorderColor());
  }


  @Test
  public void testColorFullModel() {
    ModelComparator comparator = new ModelComparator();
    Model model = createSimpleModel();
    GenericProtein protein = createProteinWithLayout();
    protein.setElementId("sa14");
    protein.setName("CNC");
    model.addElement(protein);

    Model model2 = copyModel(model);

    Model coloredModel = new CopyCommand(model).execute();

    List<DataOverlayEntry> schemas = new ArrayList<>();
    DataOverlayEntry schema = new GenericDataOverlayEntry();
    schema.setName("CNC");
    schema.setValue(-1.0);
    schemas.add(schema);
    ColorModelCommand factory = new ColorModelCommand(coloredModel, schemas, colorExtractor);

    assertNotEquals(Color.RED, coloredModel.getElementByElementId("sa14").getFillColor());

    factory.execute();

    assertEquals(0, comparator.compare(model, model2));

    assertNotEquals(0, comparator.compare(model, coloredModel));

    assertEquals(Color.RED, coloredModel.getElementByElementId("sa14").getFillColor());
  }

  private Model copyModel(final Model model) {
    return new CopyCommand(model).execute();
  }

  @Test
  public void testColoring2() {
    Model model = createSimpleModel();
    GenericProtein protein = createProteinWithLayout();
    protein.addMiriamData(new MiriamData(MiriamType.HGNC, "11138"));
    model.addElement(protein);

    Gene gene = createGene();
    gene.addMiriamData(new MiriamData(MiriamType.CHEBI, "CHEBI:15377"));
    model.addElement(gene);

    Collection<DataOverlayEntry> schemas = new ArrayList<>();
    DataOverlayEntry schema = new GenericDataOverlayEntry();
    schema.addMiriamData(new MiriamData(MiriamType.HGNC, "11138"));
    schema.setValue(1.0);
    schemas.add(schema);
    schema = new GenericDataOverlayEntry();
    schema.addMiriamData(new MiriamData(MiriamType.CHEBI, "CHEBI:15377"));
    schema.setValue(1.0);
    schemas.add(schema);
    schema = new GenericDataOverlayEntry();
    schema.addMiriamData(new MiriamData(MiriamType.CHEBI, "CHEBI:15376"));
    schema.setValue(1.0);
    schemas.add(schema);

    ColorModelCommand factory = new ColorModelCommand(model, schemas, colorExtractor);
    Collection<DataOverlayEntry> missing = factory.getMissingSchema();

    assertEquals(1, missing.size());
  }

  @Test
  public void testColorTheSameElementTwiceUsingDifferentSelector() {
    Model model = createSimpleModel();
    GenericProtein protein = createProteinWithLayout();
    protein.setName("SNCA");
    protein.addMiriamData(new MiriamData(MiriamType.HGNC, "11138"));
    model.addElement(protein);

    Collection<DataOverlayEntry> schemas = new ArrayList<>();
    DataOverlayEntry schema = new GenericDataOverlayEntry();
    schema.addMiriamData(new MiriamData(MiriamType.HGNC, "11138"));
    schema.setValue(1.0);
    schemas.add(schema);
    schema = new GenericDataOverlayEntry();
    schema.setName("SNCA");
    schema.setValue(1.0);
    schemas.add(schema);

    ColorModelCommand factory = new ColorModelCommand(model, schemas, colorExtractor);
    Map<BioEntity, DataOverlayEntry> modifiedElements = factory.getModifiedElements();

    assertEquals(1, modifiedElements.size());
  }

  @Test
  public void testReactionColoring1() {
    Model model = createSimpleModel();
    Protein protein1 = createProteinWithLayout();
    Protein protein2 = createProteinWithLayout();
    model.addElement(protein1);
    model.addElement(protein2);
    Reaction reaction = createReactionWithLayout(protein1, protein1);
    reaction.getNodes().get(0).getLine().setColor(Color.RED);
    model.addReaction(reaction);

    Collection<DataOverlayEntry> schemas = new ArrayList<>();

    ColorModelCommand factory = new ColorModelCommand(model, schemas, colorExtractor);
    assertNotEquals(Color.BLACK, reaction.getNodes().get(0).getLine().getColor());

    factory.execute();
    reaction = model.getReactionByReactionId(reaction.getIdReaction());

    assertEquals(Color.BLACK, reaction.getNodes().get(0).getLine().getColor());
  }

  @Test
  public void testReactionColoring2() {
    Model model = createSimpleModel();
    Protein protein1 = createProteinWithLayout();
    Protein protein2 = createProteinWithLayout();
    model.addElement(protein1);
    model.addElement(protein2);
    Reaction reaction = createReactionWithLayout(protein1, protein1);
    model.addReaction(reaction);

    Collection<DataOverlayEntry> schemas = new ArrayList<>();
    DataOverlayEntry schema = new GenericDataOverlayEntry();
    schema.setElementId(reaction.getIdReaction());
    schema.setColor(Color.RED);
    schemas.add(schema);

    ColorModelCommand factory = new ColorModelCommand(model, schemas, colorExtractor);
    assertEquals(Color.BLACK, reaction.getNodes().get(0).getLine().getColor());

    factory.execute();
    reaction = model.getReactionByReactionId(reaction.getIdReaction());

    assertEquals(Color.RED, reaction.getNodes().get(0).getLine().getColor());
  }

  @Test
  public void testReactionColoring3() {
    Model model = createSimpleModel();
    Protein protein1 = createProteinWithLayout();
    Protein protein2 = createProteinWithLayout();
    model.addElement(protein1);
    model.addElement(protein2);
    Reaction reaction = createReactionWithLayout(protein1, protein1);
    model.addReaction(reaction);

    Collection<DataOverlayEntry> schemas = new ArrayList<>();
    DataOverlayEntry schema = new GenericDataOverlayEntry();
    schema.setElementId(reaction.getIdReaction());
    schema.setValue(-1.0);
    schemas.add(schema);

    ColorModelCommand factory = new ColorModelCommand(model, schemas, colorExtractor);
    assertEquals(Color.BLACK, reaction.getNodes().get(0).getLine().getColor());

    factory.execute();
    reaction = model.getReactionByReactionId(reaction.getIdReaction());

    assertEquals(Color.RED, reaction.getNodes().get(0).getLine().getColor());
  }

  @Test
  public void testReactionColoring4() {
    Model model = createSimpleModel();
    Protein protein1 = createProteinWithLayout();
    Protein protein2 = createProteinWithLayout();
    model.addElement(protein1);
    model.addElement(protein2);
    Reaction reaction = createReactionWithLayout(protein1, protein1);
    reaction.addMiriamData(new MiriamData(MiriamType.PUBMED, "12345"));
    model.addReaction(reaction);

    Collection<DataOverlayEntry> schemas = new ArrayList<>();
    DataOverlayEntry schema = new GenericDataOverlayEntry();
    schema.addMiriamData(new MiriamData(MiriamType.PUBMED, "12345"));
    schema.setValue(-1.0);
    schemas.add(schema);

    ColorModelCommand factory = new ColorModelCommand(model, schemas, colorExtractor);

    assertEquals(Color.BLACK, reaction.getNodes().get(0).getLine().getColor());

    factory.execute();
    reaction = model.getReactionByReactionId(reaction.getIdReaction());

    assertEquals(Color.RED, reaction.getNodes().get(0).getLine().getColor());
  }

  @Test
  public void testColoringComplexModel() {
    Model model = createSimpleModel();
    Model model2 = createSimpleModel();

    Protein protein1 = createProteinWithLayout();
    protein1.setFillColor(Color.PINK);
    model2.addElement(protein1);

    model.addSubmodelConnection(new ModelSubmodelConnection(model2, SubmodelType.UNKNOWN, "BLA"));

    Collection<DataOverlayEntry> schemas = new ArrayList<>();
    Model coloredModel = new CopyCommand(model).execute();
    new ColorModelCommand(coloredModel, schemas, colorExtractor).execute();

    Model coloredModel2 = coloredModel.getSubmodelConnections().iterator().next().getSubmodel().getModel();
    Model coloredModel3 = coloredModel.getSubmodelByConnectionName("BLA");

    assertNotEquals(coloredModel2.getElementByElementId(protein1.getElementId()).getFillColor(),
        model2.getElementByElementId(protein1.getElementId()).getFillColor());
    assertNotEquals(coloredModel3.getElementByElementId(protein1.getElementId()).getFillColor(),
        model2.getElementByElementId(protein1.getElementId()).getFillColor());
  }

  @Test
  public void testColoredAliases() {
    Model model = createSimpleModel();
    Protein p1 = createProtein();
    p1.setName("CNC");
    model.addElement(p1);
    Protein protein = createProtein();
    protein.setName("BDH1");
    model.addElement(protein);

    Collection<DataOverlayEntry> schemas = new ArrayList<>();
    DataOverlayEntry schema = new GenericDataOverlayEntry();
    schema.setName("CNC");
    schema.setColor(Color.BLUE);
    schemas.add(schema);
    schema = new GenericDataOverlayEntry();
    schema.setName("BDH1");
    schema.setColor(Color.BLUE);
    schemas.add(schema);
    ColorModelCommand factory = new ColorModelCommand(model, schemas, colorExtractor);

    Map<BioEntity, DataOverlayEntry> map = factory.getModifiedElements();
    assertEquals(2, map.size());
    for (Map.Entry<BioEntity, DataOverlayEntry> entry : map.entrySet()) {
      assertTrue(entry.getKey() instanceof Element);
    }
  }

  @Test
  public void testColoredReactions() {
    Model model = createSimpleModel();
    Protein protein1 = createProteinWithLayout();
    Protein protein2 = createProteinWithLayout();
    model.addElement(protein1);
    model.addElement(protein2);
    Reaction reaction = createReactionWithLayout(protein1, protein1);
    model.addReaction(reaction);

    Collection<DataOverlayEntry> schemas = new ArrayList<>();
    DataOverlayEntry schema = new GenericDataOverlayEntry();
    schema.setElementId(reaction.getIdReaction());
    schema.setLineWidth(3.0);
    schema.setColor(Color.BLUE);
    schema.setName("not matching name");
    schema.setReverseReaction(true);
    schemas.add(schema);
    ColorModelCommand factory = new ColorModelCommand(model, schemas, colorExtractor);

    Map<BioEntity, DataOverlayEntry> map = factory.getModifiedElements();
    assertEquals(0, map.size());
  }

  @Test
  public void testColoredReactions2() {
    Model model = createSimpleModel();
    Protein protein1 = createProteinWithLayout();
    Protein protein2 = createProteinWithLayout();
    model.addElement(protein1);
    model.addElement(protein2);
    Reaction reaction = createReactionWithLayout(protein1, protein1);
    model.addReaction(reaction);

    Collection<DataOverlayEntry> schemas = new ArrayList<>();
    DataOverlayEntry schema = new GenericDataOverlayEntry();
    schema.setElementId(reaction.getIdReaction());
    schema.setLineWidth(3.0);
    schema.setColor(Color.BLUE);
    schema.setName(null);
    schema.setReverseReaction(true);
    schemas.add(schema);
    ColorModelCommand factory = new ColorModelCommand(model, schemas, colorExtractor);

    Map<BioEntity, DataOverlayEntry> map = factory.getModifiedElements();
    assertEquals(1, map.size());
    for (Map.Entry<BioEntity, DataOverlayEntry> entry : map.entrySet()) {
      assertTrue(entry.getKey() instanceof Reaction);
      assertEquals(reaction.getIdReaction(), ((Reaction) entry.getKey()).getIdReaction());
      assertEquals(entry.getValue(), schema);
    }
  }

  @Test
  public void testReactionColoringWithModelNotMatching() {
    Model model = createSimpleModel();
    Protein protein1 = createProteinWithLayout();
    Protein protein2 = createProteinWithLayout();
    model.addElement(protein1);
    model.addElement(protein2);
    Reaction reaction = createReactionWithLayout(protein1, protein1);
    model.addReaction(reaction);

    DataOverlayEntry schema = new GenericDataOverlayEntry();
    schema.setElementId(reaction.getIdReaction());
    schema.setName(null);
    schema.setModelName(model.getName() + "XXX");

    Collection<DataOverlayEntry> schemas = new ArrayList<>();
    schemas.add(schema);

    ColorModelCommand factory = new ColorModelCommand(model, schemas, colorExtractor);
    Map<BioEntity, DataOverlayEntry> map = factory.getModifiedElements();
    assertEquals(0, map.size());
  }

  @Test
  public void testReactionColoringWithModelMatch() {
    Model model = createSimpleModel();
    Protein protein1 = createProteinWithLayout();
    Protein protein2 = createProteinWithLayout();
    model.addElement(protein1);
    model.addElement(protein2);
    Reaction reaction = createReactionWithLayout(protein1, protein1);
    model.addReaction(reaction);

    DataOverlayEntry schema = new GenericDataOverlayEntry();
    schema.setElementId(reaction.getIdReaction());
    schema.setName(null);
    schema.setModelName(model.getName());

    Collection<DataOverlayEntry> schemas = new ArrayList<>();
    schemas.add(schema);

    ColorModelCommand factory = new ColorModelCommand(model, schemas, colorExtractor);
    Map<BioEntity, DataOverlayEntry> map = factory.getModifiedElements();
    assertEquals(1, map.size());
  }

  @Test
  public void testAliasColoringWithModelNotMatching() {
    Model model = createSimpleModel();
    Protein p1 = createProtein();
    p1.setName("CNC");
    model.addElement(p1);

    DataOverlayEntry schema = new GenericDataOverlayEntry();
    schema.setName("CNC");
    schema.setModelName(model.getName() + "XXX");

    Collection<DataOverlayEntry> schemas = new ArrayList<>();
    schemas.add(schema);

    ColorModelCommand factory = new ColorModelCommand(model, schemas, colorExtractor);
    Map<BioEntity, DataOverlayEntry> map = factory.getModifiedElements();
    assertEquals(0, map.size());
  }

  @Test
  public void testAliasColoringWithModelMatch() {
    Model model = createSimpleModel();
    Protein p1 = createProtein();
    p1.setName("CNC");
    model.addElement(p1);

    DataOverlayEntry schema = new GenericDataOverlayEntry();
    schema.setName("CNC");
    schema.setModelName(model.getName());

    Collection<DataOverlayEntry> schemas = new ArrayList<>();
    schemas.add(schema);

    ColorModelCommand factory = new ColorModelCommand(model, schemas, colorExtractor);
    Map<BioEntity, DataOverlayEntry> map = factory.getModifiedElements();
    assertEquals(1, map.size());
  }

  @Test
  public void testAliasColoringWithUnknownElementSourceId() {
    Model model = createSimpleModel();
    model.addElement(createProtein());

    DataOverlayEntry schema = new GenericDataOverlayEntry();
    schema.setElementId("xxx");

    Collection<DataOverlayEntry> schemas = new ArrayList<>();
    schemas.add(schema);

    ColorModelCommand factory = new ColorModelCommand(model, schemas, colorExtractor);
    Map<BioEntity, DataOverlayEntry> map = factory.getModifiedElements();
    assertEquals(0, map.size());
  }

  @Test
  public void testAliasColoringWithElementIdMatch() {
    Model model = createSimpleModel();
    model.addElement(createProtein());

    DataOverlayEntry schema = new GenericDataOverlayEntry();
    schema.setElementId(model.getElements().iterator().next().getElementId());

    Collection<DataOverlayEntry> schemas = new ArrayList<>();
    schemas.add(schema);

    ColorModelCommand factory = new ColorModelCommand(model, schemas, colorExtractor);
    Map<BioEntity, DataOverlayEntry> map = factory.getModifiedElements();
    assertEquals(1, map.size());
  }

}
